#!/usr/bin/python3

__author__ = "Christian O'Reilly"

import os.path
import pickle
from warnings import warn

import numpy as np
from PyQt5.QtCore import (pyqtSignal, pyqtSlot, QRect, QTimer, QSize, QObject,
                          QThread, Qt)
from PyQt5.QtGui import QPalette, QPixmap, QImage, QKeySequence
from PyQt5.QtWidgets import (QRubberBand, QLabel, QMenu, QMenuBar, QAction,
                             QVBoxLayout, QScrollArea, QSizePolicy,
                             QHBoxLayout, QPushButton, QWidget, QDialog)
from wand.image import Image


class SplashDlg(QDialog):

    def __init__(self, parent=None):
        super().__init__(parent)
        self.setWindowFlags(Qt.SplashScreen | Qt.WindowStaysOnTopHint)
        layout = QVBoxLayout()
        layout.addWidget(QLabel("Please wait while rendering PDF pages...", self))
        self.setLayout(layout)


class RenderingThread(QThread):

    def __init__(self, parent=None):
        super().__init__(parent)
        self._parent = parent
        self.renderingFinished     = False

    def run(self):
        self._parent.pages = []
        with Image(filename=self._parent.fileName, resolution=self._parent.resolution) as pdf:
            pages = len(pdf.sequence)
            for i in range(pages):
                with Image(width=pdf.width, height=pdf.height) as im:
                    im.composite(pdf.sequence[i], top=0, left=0)
                    self._parent.pages.append(im.make_blob('png'))



class PDFAreaSelector(QObject):

    areaSelected = pyqtSignal()

    def exec_(self):
        self.open()

    def __init__(self, fileName, parent=None):
        super().__init__(parent)

        self.fileName = fileName
        self.resolution = 150
        self.pages = []
        self.currentPageInd = 0

    def open(self, interactive=True):
        self.isInteractive = interactive
        if self.hasBeenCached():
            self.loadCachedRendering()
            if not self.pages is None:
                # Loading cached version succeeded
                if interactive:
                    self.selectDlg = PDFAreaSelectorDlg(self)
                    self.selectDlg.exec_()
                return

        self.waitWidget = SplashDlg()
        self.waitWidget.show()

        self.renderThread = RenderingThread(self)
        self.renderThread.start()
        self.renderThread.finished.connect(self.pdfRendered)


    @pyqtSlot()
    def pdfRendered(self):
        
        self.waitWidget.close()
        if self.isInteractive:
            self.selectDlg = PDFAreaSelectorDlg(self)
            self.selectDlg.exec_()
        self.cacheRendering()        


    def cacheRendering(self):
        fileName = self.fileName + "_cachedRendering"
        try:
            with open(fileName, 'wb') as f:
                pickle.dump(self.pages, f)
        except:
            warn("Failed to cach the rendering of the PDF paper.")
            pass


    def loadCachedRendering(self):
        fileName = self.fileName + "_cachedRendering"
        try:
            with open(fileName, 'rb') as f:
                self.pages = pickle.load(f)
        except:
            warn("Failed to load the cached rendering of the PDF paper.")
            self.pages = None



    def hasBeenCached(self):
        return os.path.isfile(self.fileName + "_cachedRendering") 











    

class PDFAreaSelectorDlg(QDialog):

    def __init__(self, parent=None):
        super().__init__(parent)

        self._parent = parent

        self.setWindowTitle("Area selection")
        self.scaleFactor = 1
        self.setModal(True)

        self.controlBarWgt         = QWidget(self)
        self.nextPageBtn         = QPushButton("Next page (Ctrl+right arrow)")
        self.previousPageBtn     = QPushButton("Previous page (Ctrl+left arrow)")
        self.noPageTxt             = QLabel("1")
        self.noPageTxt.setStyleSheet("border: 1px solid grey")
        self.noPageTxt.setFixedWidth(40)

        self.controlBarWgt.setLayout(QHBoxLayout())

        self.nextPageBtn.clicked.connect(self.nextPage)
        self.previousPageBtn.clicked.connect(self.previousPage)

        self.controlBarWgt.layout().addWidget(self.previousPageBtn)
        self.controlBarWgt.layout().addWidget(self.noPageTxt)
        self.controlBarWgt.layout().addWidget(self.nextPageBtn)

        self.imageLabel = ImageWidget()
        self.imageLabel.setBackgroundRole(QPalette.Base)
        self.imageLabel.setSizePolicy(QSizePolicy.Ignored, QSizePolicy.Ignored)
        self.imageLabel.setScaledContents(True)

        self.scrollArea = QScrollArea()
        self.scrollArea.setBackgroundRole(QPalette.Dark)
        self.scrollArea.setWidget(self.imageLabel)

        self.createActions()
        self.createMenus()

        self.setLayout(QVBoxLayout())
        self.layout().addWidget(self.menuBar)
        self.layout().addWidget(self.controlBarWgt)
        self.layout().addWidget(self.scrollArea)

        self.imageLabel.areaSelected.connect(self.resendSelectedEvent)

        self.loadImage()

    @pyqtSlot(float, float, float, float, QPixmap)
    def resendSelectedEvent(self, x, y, width, height, image):
        self._parent.x = x
        self._parent.y = y
        self._parent.width = width
        self._parent.height = height
        self._parent.image = image

        self._parent.areaSelected.emit()
        self.close()


    def loadImage(self):
        image = QImage.fromData(self._parent.pages[self._parent.currentPageInd],"PNG")

        self.imageLabel.setPixmap(QPixmap.fromImage(image))
        self.fitToWindowAct.setEnabled(True)
        self.updateActions()

        if not self.fitToWindowAct.isChecked():
            self.imageLabel.adjustSize()

        self.setWindowFilePath(self._parent.fileName)
        return True


    def zoomIn(self):
        self.scaleImage(1.25)


    def zoomOut(self):
        self.scaleImage(0.8)


    def normalSize(self):
        self.imageLabel.adjustSize()
        self.scaleFactor = 1.0

    def fitToWindow(self):
        fitToWindow = self.fitToWindowAct.isChecked()
        self.scrollArea.setWidgetResizable(fitToWindow)
        if not fitToWindow :
            self.normalSize()
        self.updateActions()



    def createActions(self):

        self.exitAct = QAction("E&xit", self)
        self.exitAct.setShortcut("Ctrl+Q")
        self.exitAct.triggered.connect(self.close)

        self.zoomInAct = QAction("Zoom &In (25%)", self)
        self.zoomInAct.setShortcut("Ctrl++")
        self.zoomInAct.setEnabled(False)
        self.zoomInAct.triggered.connect(self.zoomIn)

        self.zoomOutAct = QAction("Zoom &Out (25%)", self)
        self.zoomOutAct.setShortcut("Ctrl+-")
        self.zoomOutAct.setEnabled(False)
        self.zoomOutAct.triggered.connect(self.zoomOut)

        self.normalSizeAct = QAction("&Normal Size", self)
        self.normalSizeAct.setShortcut("Ctrl+S")
        self.normalSizeAct.setEnabled(False)
        self.normalSizeAct.triggered.connect(self.normalSize)

        self.fitToWindowAct = QAction("&Fit to Window", self)
        self.fitToWindowAct.setEnabled(False)
        self.fitToWindowAct.setCheckable(True)
        self.fitToWindowAct.setShortcut("Ctrl+F")
        self.fitToWindowAct.triggered.connect(self.fitToWindow)



        self.nextPageAct = QAction("&Next page", self)
        self.nextPageAct.setEnabled(True)
        self.nextPageAct.setShortcut(QKeySequence(Qt.CTRL + Qt.Key_Right))
        self.nextPageAct.triggered.connect(self.nextPage)


        self.previousPageAct = QAction("&Previous page", self)
        self.previousPageAct.setEnabled(True)
        self.previousPageAct.setShortcut(QKeySequence(Qt.CTRL + Qt.Key_Left))
        self.previousPageAct.triggered.connect(self.previousPage)



    def createMenus(self):

        self.menuBar = QMenuBar(self)

        self.fileMenu = QMenu("&File", self)
        self.fileMenu.addAction(self.exitAct)

        self.viewMenu = QMenu("&View", self)
        self.viewMenu.addAction(self.zoomInAct)
        self.viewMenu.addAction(self.zoomOutAct)
        self.viewMenu.addAction(self.normalSizeAct)
        self.viewMenu.addSeparator()
        self.viewMenu.addAction(self.fitToWindowAct)
        self.viewMenu.addSeparator()
        self.viewMenu.addAction(self.nextPageAct)
        self.viewMenu.addAction(self.previousPageAct)

        self.menuBar.addMenu(self.fileMenu)
        self.menuBar.addMenu(self.viewMenu)



    def updateActions(self):

        self.zoomInAct.setEnabled(not self.fitToWindowAct.isChecked())
        self.zoomOutAct.setEnabled(not self.fitToWindowAct.isChecked())
        self.normalSizeAct.setEnabled(not self.fitToWindowAct.isChecked())



    def scaleImage(self, factor):
        assert(self.imageLabel.pixmap())
        self.scaleFactor *= factor
        self.imageLabel.resize(self.scaleFactor * self.imageLabel.pixmap().size())

        self.adjustScrollBar(self.scrollArea.horizontalScrollBar(), factor)
        self.adjustScrollBar(self.scrollArea.verticalScrollBar(), factor)

        self.zoomInAct.setEnabled(self.scaleFactor < 3.0)
        self.zoomOutAct.setEnabled(self.scaleFactor > 0.333)


    def adjustScrollBar(self, scrollBar, factor):
        scrollBar.setValue(int(factor * scrollBar.value()
                                + ((factor - 1) * scrollBar.pageStep()/2)))



    def nextPage(self):
        if self._parent.currentPageInd < len(self._parent.pages)-1:
            self._parent.currentPageInd += 1
            self.loadImage()
            self.noPageTxt.setText(str(self._parent.currentPageInd+1))

    def previousPage(self):
        if self._parent.currentPageInd > 0:
            self._parent.currentPageInd -= 1
            self.loadImage()
            self.noPageTxt.setText(str(self._parent.currentPageInd+1))






class ImageWidget(QLabel):

    areaSelected = pyqtSignal(float, float, float, float, QPixmap)

    def __init__(self, parent=None):
        super().__init__(parent)
        self.rubberBand  = None
        self.origin         = None

    def mousePressEvent(self, event):
        self.origin = event.pos()
        if self.rubberBand is None:
            self.rubberBand = QRubberBand(QRubberBand.Rectangle, self)
        self.rubberBand.setGeometry(QRect(self.origin, QSize()))
        self.rubberBand.show()

    def mouseMoveEvent(self, event):
        self.rubberBand.setGeometry(QRect(self.origin, event.pos()).normalized())


    def mouseReleaseEvent(self, event):
        self.rubberBand.hide()
        zoreRect = QRect(self.origin, event.pos()).normalized()
        pixmapSize = self.pixmap().size()
        pixX1 = zoreRect.x()
        pixX2 = zoreRect.x() + zoreRect.width()
        pixY1 = zoreRect.y()
        pixY2 = zoreRect.y() + zoreRect.height()
        width  = pixmapSize.width()
        height = pixmapSize.height()

        x1 = pixX1/width 
        x1 = x1 if x1 >= 0.0 else 0.0
        x1 = x1 if x1 <= 1.0 else 1.0
        y1 = pixY1/height 
        y1 = y1 if y1 >= 0.0 else 0.0
        y1 = y1 if y1 <= 1.0 else 1.0
        x2 = pixX2/width 
        x2 = x2 if x2 >= 0.0 else 0.0
        x2 = x2 if x2 <= 1.0 else 1.0
        y2 = pixY2/height 
        y2 = y2 if y2 >= 0.0 else 0.0
        y2 = y2 if y2 <= 1.0 else 1.0

        rect = QRect(min(pixX1, pixX2), min(pixY1, pixY2), abs(pixX1- pixX2), abs(pixY1- pixY2))
        selectedImg = self.pixmap().copy(rect)
        
        self.areaSelected.emit(min(x1, x2), min(y1, y2), np.abs(x1-x2), np.abs(y1-y2), selectedImg)
    









def loadImage(fileName, pageNo, x, y, width, height):

    areaSelector = PDFAreaSelector(fileName)
    areaSelector.open(interactive=False)
    i = 0
    timer = QTimer()
    while len(areaSelector.pages) == 0 and i < 120:
        i += 1
        timer.start(500)
        
    image = QImage.fromData(areaSelector.pages[pageNo-1],"PNG")
    pixmap = QPixmap.fromImage(image)

    pixmapSize = pixmap.size()
    widthImg  = pixmapSize.width()
    heightImg = pixmapSize.height()

    rect = QRect(x*widthImg, y*heightImg, width*widthImg, height*heightImg)
    return pixmap.copy(rect)
    










from sys import platform as _platform

from PyQt5.QtCore import (pyqtSignal, QSortFilterProxyModel, QStringListModel,
                          QEvent, QRegExp, Qt)
from PyQt5.QtWidgets import QCompleter, QComboBox


class CustomQCompleter(QCompleter):
    """
    adapted from: http://stackoverflow.com/a/7767999/2156909
    """

    def __init__(self, parent=None):
        super().__init__(parent)
        self.local_completion_prefix = ""
        self.source_model = None
        self.filterProxyModel = QSortFilterProxyModel(self)
        self.usingOriginalModel = False

    def setModel(self, strList): 
        self.source_model = QStringListModel(strList)
        self.filterProxyModel = QSortFilterProxyModel(self)
        self.filterProxyModel.setSourceModel(self.source_model)
        super().setModel(self.filterProxyModel)
        self.usingOriginalModel = True


    def updateModel(self):
        if not self.usingOriginalModel:
            self.filterProxyModel.setSourceModel(self.source_model)

        pattern = QRegExp(self.local_completion_prefix,
                                Qt.CaseInsensitive,
                                QRegExp.FixedString)

        self.filterProxyModel.setFilterRegExp(pattern)

    def splitPath(self, path):
        self.local_completion_prefix = path
        self.updateModel()
        if self.filterProxyModel.rowCount() == 0:
            self.usingOriginalModel = False
            self.filterProxyModel.setSourceModel(QStringListModel([path]))
            return [path]

        self.usingOriginalModel = True
        return []

class AutoCompleteEdit(QComboBox):

    enterKeyPressed = pyqtSignal(QComboBox)

    def __init__(self, parent=None):
        super().__init__(parent)
        self.setEditable(True)
        self.setInsertPolicy(self.NoInsert)
        self.comp = CustomQCompleter(self)
        self.comp.setCompletionMode(QCompleter.PopupCompletion)
        self.setCompleter(self.comp)
        self.setEditText("")
        self.erase = False
        self.deactivateClearing = False

    def setModel(self, strList):
        self.clear()
        self.insertItems(0, strList)
        self.comp.setModel(strList)

    def focusInEvent(self, event):
        
    
        if _platform in ["linux", "linux2"]:
              # This behavior is the original one, which is fine in linux. On MacOS,
            # selecting an item change the value of the combobox and then set the focus on it
            # which was making this line erase the selected text.
            if not self.deactivateClearing :
                self.clearEditText()

        super().focusInEvent(event)


    def event(self, event):
        if event.type() == QEvent.KeyPress:
            if event.key() == Qt.Key_Enter or event.key() == 16777220:
                # Enter (if event.key() == Qt.Key_Enter) does not work
                # for some reason
 
                self.deactivateClearing = True
                # make sure that the completer does not set the
                # currentText of the combobox to "" when pressing enter
                text = self.currentText()
                self.setCompleter(None)
                self.setEditText(text)
                self.setCompleter(self.comp)
                self.enterKeyPressed.emit(self)
                self.deactivateClearing = False
                
        eventReturn = super().event(event)
        return eventReturn



__authors__ = ["Pierre-Alexandre Fonta", "Christian O'Reilly"]
__maintainer__ = "Pierre-Alexandre Fonta"

from PyQt5.QtCore import QThread


class ZoteroRefreshThread(QThread):

    def __init__(self, zotero_model, parent=None):
        super().__init__(parent)
        # NB: Executes in the old thread.
        self._zotero_model = zotero_model

    def run(self):
        # NB: Executes in the new thread. Beware of accessing __init__() variables.
        # NB: Returning from this method will end the execution of the thread.
        # QMutex/QMutexLocker not needed if ZoteroTableModel is not modified.
        # QThread::exec() not needed if an event loop inside the thread is not needed.
        self._zotero_model.refresh()



#!/usr/bin/python3

__author__ = "Christian O'Reilly"

from PyQt5.QtCore import QModelIndex, QAbstractTableModel, Qt


class PropositionTableModel(QAbstractTableModel):

    def __init__(self, parent=None):
        super().__init__(parent)
        self.header = ["value", "unit", "authors", "year", "journal"]
        self.nbCol = len(self.header)
        self.propositions = []

    def refreshData(self, annotatedInstances, modelingInstance):
        self.propositions = []

        for annotInstance in annotatedInstances:
            proposition = {"value": annotInstance.value,
                           "unit": annotInstance.unit,
                           "authors": "",
                           "year": "",
                           "journal": "",
                           "parameterInstance": annotInstance}
            self.propositions.append(proposition)

        self.layoutChanged.emit()

    def rowCount(self, parent=QModelIndex()):
        return len(self.propositions)

    def columnCount(self, parent=QModelIndex()):
        return self.nbCol 



    def data(self, index, role=Qt.DisplayRole):
        if not index.isValid():
            return None


        #if role == Qt.BackgroundRole:
        #    if self.checkIdFct(self.getID(index.row())):
        #        return QtGui.QBrush(QtGui.QColor(215, 214, 213), Qt.SolidPattern)
        #    else:
        #        return None

        if role == Qt.DisplayRole:
            try:
                return self.propositions[index.row()][self.header[index.column()]]
            except KeyError:
                return ""
        return None


    def headerData(self, col, orientation, role=Qt.DisplayRole):
        if orientation == Qt.Horizontal and role == Qt.DisplayRole:
            return self.header[col]
        return None

    #def sort(self, col, order):
    #    #sort table by given column number col
    #    self.layoutAboutToBeChanged.emit()
    #    reverse = (order == Qt.DescendingOrder)
    #    self.refList = sorted(self.refList, key=lambda x: self.getByIndex(x, col), reverse = reverse)
    #    self.layoutChanged.emit()

    def refresh(self):
        self.layoutChanged.emit()



#!/usr/bin/python3

__author__ = "Christian O'Reilly"

from PyQt5.QtCore import (QModelIndex, pyqtSignal, pyqtSlot, Qt, QAbstractTableModel)
from PyQt5.QtWidgets import (QTableView, QMessageBox, QLabel, QGridLayout,
                             QTabWidget, QLineEdit, QAbstractItemView, QWidget,
                             QPushButton)

from nat.annotation import getParametersForPub
from nat.modelingParameter import (getParameterTypes, ParameterTypeTree,
                                   getParameterTypeNameFromID)
from nat.paramDesc import ParamDescFunction, InvalidEquation, ParamRef
from nat.parameterInstance import ParameterInstance
from nat.utils import Id2FileName
from .itemDelegates import CheckBoxDelegate
from .variableTableWgt import VariableTableView, VariableListModel


parameterTypeTree = ParameterTypeTree.load()
parameterTypes = getParameterTypes()


class ParamFunctionWgt(QWidget):

    paramTypeSelected = pyqtSignal(str)
    
    def __init__(self, parent=None):
        super().__init__(parent)

        self._parent = parent

        self.parameterTypes = getParameterTypes()

        # Widgets        
        self.addIndepVarBtn = QPushButton("Add variable")
        self.deleteIndepVarBtn = QPushButton("Delete variable")
        self.varListTblWdg = VariableTableView()
        self.varListModel = VariableListModel(parent=self)
        self.varListTblWdg.setSelectionMode(QAbstractItemView.SingleSelection)
        self.varListTblWdg.setModel(self.varListModel)

        self.varTab                = QWidget(self)
        grid                       = QGridLayout(self.varTab)
        grid.addWidget(self.varListTblWdg, 0, 0, 3, 1)
        grid.addWidget(self.addIndepVarBtn, 0, 1)
        grid.addWidget(self.deleteIndepVarBtn, 1, 1)

        self.paramListTblWdg       = ParameterInstanceTableView()
        self.paramListModel        = ParameterInstanceListModel(parent=self)
        self.paramListTblWdg.setSelectionBehavior(QAbstractItemView.SelectRows)
        self.paramListTblWdg.setSelectionMode(QAbstractItemView.SingleSelection)
        self.paramListTblWdg.setModel(self.paramListModel)

        self.functionTxt        = QLineEdit(self)
        self.functionDepPartTxt    = QLabel(self)

        self.eqElementsTabs = QTabWidget(self)
        self.eqElementsTabs.addTab(self.varTab,   "Equation variables")
        self.eqElementsTabs.addTab(self.paramListTblWdg, "Equation parameters")
        #self.eqElementsTabs.setTabEnabled(1, False)
        
        #TODO: We need to implement some kind of search to allow the user to select an existing parameter.
        # It need not be from the same annotation, nor even the same paper (in case they took a parameter 
        # values from a paper they cite).

        # Signals
        self.addIndepVarBtn.clicked.connect(self.varListModel.addVariable)
        self.deleteIndepVarBtn.clicked.connect(self.varListTblWdg.deleteVariable)
        self.varListTblWdg.depTypeSelected.connect(self.setDepText)

        # Layout
        grid     = QGridLayout(self)

        grid.addWidget(self.eqElementsTabs, 0, 0, 1, 3)
        grid.addWidget(QLabel("Equation:  "), 2, 0)
        grid.addWidget(self.functionDepPartTxt, 2, 1)
        grid.addWidget(self.functionTxt, 2, 2)

    @pyqtSlot(str)
    def setDepText(self, depVar):
        self.functionDepPartTxt.setText(depVar + " = ")
        self.paramTypeSelected.emit(depVar)

    def newParameter(self):
        self.varListModel.clear()
        self.paramListModel.clear()
        self.functionTxt.setText("")
        self.functionDepPartTxt.setText("")
        self.fillingEquationParameterList()


    def saveParameter(self, relationship, paramId):
        depVar        = self.varListModel.getDepVar(varType="Variable")
        indepVars     = self.varListModel.getIndepVars(varType="Variable")

        parameters = []
        for noRow in range(self.paramListModel.rowCount()):
            paramInstanceId = self.paramListModel.getParamInstanceID(noRow)
            if self.paramListModel.selected[paramInstanceId]:
                parameters.append(ParamRef(paramInstanceId, 
                                       self.paramListModel.getParamTypeID(noRow)))

        equation        = self.functionDepPartTxt.text() + self.functionTxt.text()

        try:
            description = ParamDescFunction(depVar, indepVars, parameters, equation)
        except InvalidEquation:
            msgBox = QMessageBox(self)
            msgBox.setWindowTitle("Invalid equation")
            msgBox.setText("The equation '" + equation + "' is not a valid Python expression. Please correct this expression and try saving again.")
            msgBox.exec_()
            return None

        return ParameterInstance(paramId, description, [], [], relationship)


    def loadRow(self, currentParameter = None):
        #self.varListModel.load(currentParameter)
        
        if currentParameter is None:
            self.functionDepPartTxt.setText("")
            self.functionTxt.setText("")
            self.varListModel.clear()
            self.paramListModel.load([], [])        
            self.varListModel.refresh()
            self.paramListModel.refresh()
            return
            

        #paramInstances = []
        #for id in currentParameter.description.parameterIds:
        #    for param in self.mainWgt.paramListModel.parameterList:
        #        if param.id == id :
        #            paramInstances.append(param)
        #            break
        #
        #self.paramListModel.load(paramInstances)
        left, right = currentParameter.description.equation.split("=")
        left  = left.strip()
        right = right.strip()
        self.functionDepPartTxt.setText(left + " = ")
        self.functionTxt.setText(right)

        self.varListModel.setFromParam(currentParameter)
        self.varListModel.refresh()

        self.fillingEquationParameterList()





    def fillingEquationParameterList(self, currentParameter = None):
        parameters = getParametersForPub(self._parent.dbPath, Id2FileName(self._parent.currentAnnotation.pubId))
        parameters = [param for param in parameters if param.isExperimentProperty == False]
        if currentParameter is None:
            selectedParams = []
        else:
            selectedParams = [ref.instanceId for ref in currentParameter.description.parameterRefs]        
        self.paramListModel.load(parameters, selectedParams)
        self.paramListModel.refresh()
        


    def loadModelingParameter(self, row = None):
        pass


class ParameterInstanceTableView(QTableView):

    tableCheckBoxClicked = pyqtSignal(int)

    def __init__(self, parent=None):
        super().__init__(parent)
        self.setItemDelegateForColumn(0, CheckBoxDelegate(self))      

    @pyqtSlot(bool)
    def checkBoxClicked(self, checked):
        # This slot will be called when our button is clicked. 
        # self.sender() returns a refence to the QPushButton created
        # by the delegate, not the delegate itself.
        #self.model().deleteRow(self.sender().row)
        self.model().selectParameter(self.sender().row, not checked)
        self.tableCheckBoxClicked.emit(self.sender().row)


class ParameterInstanceListModel(QAbstractTableModel):

    def __init__(self, colHeader = ['', 'Type', 'Description'], parent=None): #'Type',
        super().__init__(parent)
        #self.parameterList = parameterList
        self.colHeader = colHeader
        self.nbCol     = len(colHeader)
        #self.types       = []
        self.instances = []
        self.selected  = {}

    def clear(self):
        for id in self.selected:
            self.selected[id] = False
        self.refresh()

    def rowCount(self, parent=QModelIndex()):
        return len(self.instances)

    def columnCount(self, parent=QModelIndex()):
        return self.nbCol 


    def getParamInstanceID(self, row):
        return self.instances[row].id
 
    def getParamTypeID(self, row):
        return self.instances[row].description.depVar.typeId

    def getByIndex(self, param, ind):
        if ind == 0:
            return self.selected[param.id]
        elif ind == 1:
            return param.typeDesc
        elif ind == 2:
            if param.typeDesc == "pointValue":
                return param.name + " = " + param.description.depVar.values.text(True)
            elif param.typeDesc == "function":
                return param.description.equation
            elif param.typeDesc == "numericalTrace":
                return getParameterTypeNameFromID(param.description.depVar.typeId) \
                          + "=f(" + " ,".join([getParameterTypeNameFromID(v.typeId) for v in param.description.indepVars]) +  ")"
            else:
                raise ValueError
        else:
            raise ValueError


    def data(self, index, role=Qt.DisplayRole):
        if not index.isValid():
            return None

        if role != Qt.DisplayRole:
            return None

        return self.getByIndex(self.instances[index.row()], index.column())


    def flags(self, index):
        return Qt.ItemIsEditable | Qt.ItemIsEnabled



    def headerData(self, section, orientation, role=Qt.DisplayRole):
        if orientation == Qt.Horizontal and role == Qt.DisplayRole:
            return self.colHeader[section]        
        return None


    def load(self, paramInstances, selectedIDs):
        self.instances = paramInstances
        self.selected  = {}
        for row, instance in enumerate(self.instances):
            ID = self.getParamInstanceID(row)
            self.selected[ID] = ID in selectedIDs
        self.refresh()
        
    def selectParameter(self, row, selected):
        ID = self.getParamInstanceID(row)
        self.selected[ID] = selected     


    def refresh(self):
        self.layoutChanged.emit()



#!/usr/bin/python3

__author__ = "Christian O'Reilly"

import re
from collections import OrderedDict
from difflib import SequenceMatcher
from os.path import join

import numpy as np
from PyQt5.QtCore import pyqtSlot, Qt
from PyQt5.QtGui import QPalette
from PyQt5.QtWidgets import (QMessageBox, QDialog, QLabel, QGridLayout,
                             QGroupBox, QLineEdit, QPushButton, QWidget,
                             QFrame, QTextEdit, QStackedWidget, QComboBox)

from nat.annotation import (TextLocalizer, FigureLocalizer, TableLocalizer,
                            EquationLocalizer, PositionLocalizer,
                            NullLocalizer,
                            Annotation)
from nat.utils import Id2FileName
from .approximateMatchDlg import MatchDlg
from .areaSelector import PDFAreaSelector, loadImage
from .jsonDlg import JSONDlg
from .uiUtilities import disableTextWidget, enableTextWidget
from .uiUtilities import errorMessage


class EditAnnotWgt(QWidget):

    def __init__(self, parent=None):
        super().__init__(parent)

        self.main_window = parent

        # Widgets
        self.annotationTypesCbo     = QComboBox(self)
        self.newAnnotationBtn       = QPushButton('New', self)
        self.duplicateAnnotationBtn = QPushButton('Duplicate', self)
        self.deleteAnnotationBtn    = QPushButton('Delete', self)
        self.viewJSONBtn            = QPushButton('View JSON', self)
        self.saveAnnotationBtn      = QPushButton('Save', self)
        self.commentEdt             = QTextEdit('', self)

        self.editAnnotWgt = OrderedDict([("text",     EditAnnotTextWgt(self)), 
                                         ("figure",   EditAnnotFigureWgt(self)), 
                                         ("table",    EditAnnotTableWgt(self)), 
                                         ("equation", EditAnnotEquationWgt(self)), 
                                         ("position", EditAnnotPositionWgt(self)),
                                         ("null",     EditAnnotNullWgt(self))])

        # We use the list to as a redundancy of the list of keys in annotTypesDict
        # because we want these to be ordered. 
        self.annotTypeLst     = list(self.editAnnotWgt.keys())

        self.annotationTypesCbo.addItems(self.annotTypeLst)

        self.editAnnotStack = QStackedWidget(self)

        for annotType in self.annotTypeLst:
            self.editAnnotStack.addWidget(self.editAnnotWgt[annotType])

        # Layout
        self.editAnnotGroupBox = QGroupBox()
        gridAddAnnotations = QGridLayout(self.editAnnotGroupBox)
        gridAddAnnotations.addWidget(QLabel('Annotation type', self), 2, 0)
        gridAddAnnotations.addWidget(self.annotationTypesCbo, 2, 1)
        gridAddAnnotations.addWidget(self.saveAnnotationBtn, 2, 2)
        gridAddAnnotations.addWidget(self.deleteAnnotationBtn, 2, 3)
        gridAddAnnotations.addWidget(self.newAnnotationBtn, 2, 4)
        gridAddAnnotations.addWidget(self.duplicateAnnotationBtn, 2, 5)
        gridAddAnnotations.addWidget(self.viewJSONBtn, 2, 6)

        gridAddAnnotations.addWidget(self.editAnnotStack, 3, 0, 1, 7)

        gridAddAnnotations.addWidget(QLabel('Comment', self), 4, 0)
        gridAddAnnotations.addWidget(self.commentEdt, 4, 1, 1, 6)
        self.setLayout(gridAddAnnotations)

        # Signals
        self.saveAnnotationBtn.clicked.connect(self.saveAnnotation)
        self.newAnnotationBtn.clicked.connect(self.newAnnotation)
        self.duplicateAnnotationBtn.clicked.connect(self.duplicateAnnotation)
        self.viewJSONBtn.clicked.connect(self.viewJSON)
        self.deleteAnnotationBtn.clicked.connect(self.main_window.deleteAnnotation)
        self.commentEdt.textChanged.connect(self.annotationChanged)
        self.annotationTypesCbo.currentIndexChanged.connect(self.setCurrentStack)

        self.main_window.selectedAnnotationChangedConfirmed.connect(self.annotationSelectionChanged)
        for wgt in self.editAnnotWgt.values():
            self.main_window.selectedAnnotationChangedConfirmed.connect(wgt.annotationSelectionChanged)
        self.main_window.annotationCleared.connect(self.clearAddAnnotation)
        self.main_window.savingNeeded.connect(self.savingNeeded)

        # Initial behavior
        self.deleteAnnotationBtn.setDisabled(True)
        self.viewJSONBtn.setDisabled(True)
        self.annotationTypesCbo.setCurrentIndex(0)
        self.setCurrentStack(0)
        self.annotationTypesCbo.setDisabled(True)

    def viewJSON(self):        
        form = JSONDlg()
        form.setJSON(self.currentAnnotation)
        if form.exec_() == QDialog.Accepted:
            return 

    def saveAnnotation(self):
        self.main_window.saveAnnotation()
        self.annotationTypesCbo.setDisabled(True)

        
    def selectAnnotType(self, annotType):
        
        ind = [no for no, a in enumerate(self.annotTypeLst) if a == annotType]
        if len(ind) != 1:
            raise ValueError("Invalid annotation type string.")
        ind = ind[0]

        self.annotationTypesCbo.setCurrentIndex(ind)
        self.annotationTypesCbo.setEnabled(False)


    def setCurrentStack(self, ind):
        self.editAnnotStack.setCurrentIndex(ind) 

        #for no in range(len(self.annotTypeLst)):
        #    self.editAnnotStack.setTabEnabled(no, no == ind)
    


    @pyqtSlot()
    def clearAddAnnotation(self):
        self.commentEdt.setText("")
        for widget in self.editAnnotWgt.values():
            widget.clearAnnotation()

    @property
    def detectAnnotChange(self):
        return self.main_window.detectAnnotChange


    def annotationTextChanged(self):
        if self.detectAnnotChange:
            self.main_window.setNeedSaving()


    def annotationChanged(self):
        if self.detectAnnotChange:
            self.main_window.setNeedSaving()


    @property
    def currentAnnotation(self):
        return self.main_window.currentAnnotation

    @currentAnnotation.setter
    def currentAnnotation(self, annotation):
        self.main_window.currentAnnotation = annotation


    @pyqtSlot()
    def annotationSelectionChanged(self):
        #self.deleteAnnotationBtn.setEnabled(True)
        self.newAnnotationBtn.setEnabled(True)
        self.duplicateAnnotationBtn.setEnabled(True)
        if not self.currentAnnotation is None:
            self.commentEdt.setText(self.currentAnnotation.comment)
            enableTextWidget(self.commentEdt)
        else:
            self.commentEdt.setText("")
            disableTextWidget(self.commentEdt)            
        
        self.deleteAnnotationBtn.setDisabled(self.currentAnnotation is None)
        self.viewJSONBtn.setDisabled(self.currentAnnotation is None)

    def newAnnotation(self):
        if self.main_window.newAnnotation() :
            self.newAnnotationBtn.setEnabled(False)
            self.duplicateAnnotationBtn.setEnabled(False)
            self.deleteAnnotationBtn.setEnabled(False)
            self.viewJSONBtn.setDisabled(False)
            self.annotationTypesCbo.setEnabled(True)
            self.clearAddAnnotation()
            for widget in self.editAnnotWgt.values():
                widget.newAnnotation()

            if "UNPUBLISHED" in self.main_window.IdTxt.text():
                self.annotationTypesCbo.setEnabled(False)
                self.setCurrentStack(self.annotTypeLst.index("null"))
                self.annotationTypesCbo.setCurrentIndex(self.annotTypeLst.index("null"))
            else:
                self.annotationTypesCbo.setEnabled(True)
            enableTextWidget(self.commentEdt)

    def duplicateAnnotation(self):
        self.main_window.duplicateAnnotation()



    @pyqtSlot(bool)
    def savingNeeded(self, needSaving):
        self.saveAnnotationBtn.setEnabled(needSaving)
        self.updateCurrentAnnotation()



    def updateCurrentAnnotation(self):
        if not self.currentAnnotation is None:
            self.currentAnnotation.comment = self.commentEdt.toPlainText()
            if not self.main_window.username in self.currentAnnotation.users:
                self.currentAnnotation.users.append(self.main_window.username)
            self.editAnnotWgt[self.annotationTypesCbo.currentText()].updateCurrentAnnotation()



class EditAnnotFigureWgt(QWidget):

    def __init__(self, container, parent=None):
        super().__init__(parent)

        self.container = container

        # Widgets
        self.noFigure                = QLineEdit('', self)
    
        # Signals
        self.noFigure.textChanged.connect(self.container.annotationChanged)

        # Layout
        self.editAnnotGroupBox = QGroupBox()
        gridAddAnnotations = QGridLayout(self.editAnnotGroupBox)
        gridAddAnnotations.addWidget(QLabel('Figure no.', self), 2, 0)
        gridAddAnnotations.addWidget(self.noFigure, 2, 1)

        self.setLayout(gridAddAnnotations)

        # Initial behavior
        self.editAnnotGroupBox.setDisabled(True)
        disableTextWidget(self.noFigure)

    @pyqtSlot()
    def annotationSelectionChanged(self):
        if not self.container.currentAnnotation is None:
            enableTextWidget(self.noFigure)
            if self.container.currentAnnotation.type == "figure":
                self.noFigure.setText(self.container.currentAnnotation.localizer.no)
        else:
            disableTextWidget(self.noFigure)
            self.noFigure.setText("")


    def newAnnotation(self):
        enableTextWidget(self.noFigure)


    def clearAnnotation(self):
        self.noFigure.setText("")


    def updateCurrentAnnotation(self):
        self.container.currentAnnotation.localizer = FigureLocalizer(self.noFigure.text())




class EditAnnotNullWgt(QWidget):

    def __init__(self, container, parent=None):
        super().__init__(parent)
        self.container = container

    @pyqtSlot()
    def annotationSelectionChanged(self):
        pass


    def newAnnotation(self):
        pass


    def clearAnnotation(self):
        pass


    def updateCurrentAnnotation(self):
        self.container.currentAnnotation.localizer = NullLocalizer()




class EditAnnotEquationWgt(QWidget):

    def __init__(self, container, parent=None):
        super().__init__(parent)

        self.container = container

        # Widgets
        self.noEquation                = QLineEdit('', self)
        self.pythonStringEdt        = QTextEdit('', self)
    
        # Signals
        self.noEquation.textChanged.connect(self.container.annotationChanged)
        self.pythonStringEdt.textChanged.connect(self.container.annotationChanged)

        # Layout
        self.editAnnotGroupBox = QGroupBox()
        gridAddAnnotations = QGridLayout(self.editAnnotGroupBox)
        gridAddAnnotations.addWidget(QLabel('Equation no.', self), 2, 0)
        gridAddAnnotations.addWidget(self.noEquation, 2, 1)

        gridAddAnnotations.addWidget(QLabel('Python representation', self), 3, 0)
        gridAddAnnotations.addWidget(self.pythonStringEdt, 3, 1)
        gridAddAnnotations.addWidget(QLabel('(optional)', self), 3, 2)

        self.setLayout(gridAddAnnotations)

        # Initial behavior
        self.editAnnotGroupBox.setDisabled(True)
        disableTextWidget(self.noEquation)
        disableTextWidget(self.pythonStringEdt)

    @pyqtSlot()
    def annotationSelectionChanged(self):

        if not self.container.currentAnnotation is None:
            enableTextWidget(self.noEquation)
            enableTextWidget(self.pythonStringEdt)
            if self.container.currentAnnotation.type == "equation":
                self.noEquation.setText(self.container.currentAnnotation.localizer.no)
                self.pythonStringEdt.setText(self.container.currentAnnotation.localizer.equation)
        else:
            disableTextWidget(self.noEquation)
            disableTextWidget(self.pythonStringEdt)
            self.noEquation.setText("")
            self.pythonStringEdt.setText("")


    def newAnnotation(self):
        enableTextWidget(self.noEquation)
        enableTextWidget(self.pythonStringEdt)


    def clearAnnotation(self):
        self.noEquation.setText("")
        self.pythonStringEdt.setText("")


    def updateCurrentAnnotation(self):
        pythonString = self.pythonStringEdt.toPlainText() if self.pythonStringEdt.toPlainText() != "" else None
        self.container.currentAnnotation.localizer = EquationLocalizer(self.noEquation.text(), pythonString)








class EditAnnotTableWgt(QWidget):

    def __init__(self, container, parent=None):
        super().__init__(parent)

        self.container = container

        # Widgets
        self.noTable                = QLineEdit('', self)
        self.noRow                    = QLineEdit('', self)
        self.noCol                    = QLineEdit('', self)
    
        # Signals
        self.noTable.textChanged.connect(self.container.annotationChanged)
        self.noRow.textChanged.connect(self.container.annotationChanged)
        self.noCol.textChanged.connect(self.container.annotationChanged)

        # Layout
        self.editAnnotGroupBox = QGroupBox()
        gridAddAnnotations = QGridLayout(self.editAnnotGroupBox)
        gridAddAnnotations.addWidget(QLabel('Table no.', self), 2, 0)
        gridAddAnnotations.addWidget(self.noTable, 2, 1)

        gridAddAnnotations.addWidget(QLabel('Row no.', self), 3, 0)
        gridAddAnnotations.addWidget(self.noRow, 3, 1)
        gridAddAnnotations.addWidget(QLabel('(optional)', self), 3, 2)

        gridAddAnnotations.addWidget(QLabel('Column no.', self), 4, 0)
        gridAddAnnotations.addWidget(self.noCol, 4, 1)
        gridAddAnnotations.addWidget(QLabel('(optional)', self), 4, 2)

        self.setLayout(gridAddAnnotations)

        # Initial behavior
        self.editAnnotGroupBox.setDisabled(True)
        disableTextWidget(self.noTable)

    @pyqtSlot()
    def annotationSelectionChanged(self):

        if not self.container.currentAnnotation is None:
            enableTextWidget(self.noTable)
            enableTextWidget(self.noRow)
            enableTextWidget(self.noCol)
            if self.container.currentAnnotation.type == "table":
                self.noTable.setText(self.container.currentAnnotation.localizer.no)
                self.noRow.setText(str(self.container.currentAnnotation.localizer.noRow))
                self.noCol.setText(str(self.container.currentAnnotation.localizer.noCol))
        else:
            disableTextWidget(self.noTable)
            disableTextWidget(self.noRow)
            disableTextWidget(self.noCol)
            self.noTable.setText("")
            self.noRow.setText("")
            self.noCol.setText("")


    def newAnnotation(self):
        enableTextWidget(self.noTable)
        enableTextWidget(self.noRow)
        enableTextWidget(self.noCol)
        #self.noTable.setFocus()

    def clearAnnotation(self):
        self.noTable.setText("")
        self.noRow.setText("")
        self.noCol.setText("")

    def updateCurrentAnnotation(self):
        noRow = self.noRow.text() if self.noRow.text() != "" else None
        noCol = self.noCol.text() if self.noCol.text() != "" else None
        self.container.currentAnnotation.localizer = TableLocalizer(self.noTable.text(), noRow, noCol)





class ImageThumbnail(QLabel):

    def __init__(self, parent=None):
        super().__init__(parent)
        self.setBackgroundRole(QPalette.Base)
        self.setFrameStyle(QFrame.StyledPanel | QFrame.Sunken)
        #self.setScaledContents(True)

    #def resizeEvent(self, event):
    #    img = QtGui.QPixmap()
    #    w   = self.width()
    #    h   = self.height()
    #    self.setPixmap(self.pixmap().scaled(w, h, Qt.KeepAspectRatio))


class EditAnnotPositionWgt(QWidget):

    def __init__(self, container, parent=None):
        super().__init__(parent)

        self.container = container

        # Widgets
        self.noPageTxt         = QLineEdit('', self)
        self.xTxt             = QLineEdit('', self)
        self.yTxt             = QLineEdit('', self)
        self.widthTxt         = QLineEdit('', self)
        self.heightTxt         = QLineEdit('', self)
        self.imgThumbnail   = ImageThumbnail(self)
        self.selectAreaBtn     = QPushButton('Select area', self)

        # Signals
        self.noPageTxt.textChanged.connect(self.container.annotationChanged)
        self.selectAreaBtn.clicked.connect(self.selectArea)

        # Layout
        self.editAnnotGroupBox = QGroupBox()
        gridAddAnnotations = QGridLayout(self.editAnnotGroupBox)
        gridAddAnnotations.addWidget(QLabel('Page no.', self), 2, 0)
        gridAddAnnotations.addWidget(self.noPageTxt, 2, 1)
        gridAddAnnotations.addWidget(QLabel('x', self), 3, 0)
        gridAddAnnotations.addWidget(self.xTxt, 3, 1)
        gridAddAnnotations.addWidget(QLabel('y', self), 4, 0)
        gridAddAnnotations.addWidget(self.yTxt, 4, 1)
        gridAddAnnotations.addWidget(QLabel('width', self), 5, 0)
        gridAddAnnotations.addWidget(self.widthTxt, 5, 1)
        gridAddAnnotations.addWidget(QLabel('height', self), 6, 0)
        gridAddAnnotations.addWidget(self.heightTxt, 6, 1)
        gridAddAnnotations.addWidget(self.selectAreaBtn, 7, 0, 1, 2)
        gridAddAnnotations.addWidget(self.imgThumbnail, 2, 2, 6, 1)

        gridAddAnnotations.setColumnStretch(0,1)
        gridAddAnnotations.setColumnStretch(1,1)
        gridAddAnnotations.setColumnStretch(2,3)

        self.setLayout(gridAddAnnotations)

        # Initial behavior
        self.editAnnotGroupBox.setDisabled(True)
        disableTextWidget(self.noPageTxt)
        disableTextWidget(self.xTxt)
        disableTextWidget(self.yTxt)
        disableTextWidget(self.widthTxt)
        disableTextWidget(self.heightTxt)


    @pyqtSlot()
    def annotationSelectionChanged(self):

        if not self.container.currentAnnotation is None:
            if self.container.currentAnnotation.type == "position":
                self.noPageTxt.setText(str(self.container.currentAnnotation.localizer.noPage))
                self.xTxt.setText(str(self.container.currentAnnotation.localizer.x))
                self.yTxt.setText(str(self.container.currentAnnotation.localizer.y))
                self.widthTxt.setText(str(self.container.currentAnnotation.localizer.width))
                self.heightTxt.setText(str(self.container.currentAnnotation.localizer.height))
                self.loadThumbnail()
        else:
            self.noPageTxt.setText("")
            self.xTxt.setText("")
            self.yTxt.setText("")
            self.widthTxt.setText("")
            self.heightTxt.setText("")
            self.imgThumbnail.clear()

    def newAnnotation(self):
        pass


    def clearAnnotation(self):
        self.noPageTxt.setText("")
        self.xTxt.setText("")
        self.yTxt.setText("")
        self.widthTxt.setText("")
        self.heightTxt.setText("")


    def updateCurrentAnnotation(self):
        self.container.currentAnnotation.localizer = PositionLocalizer(int(self.noPageTxt.text()), float(self.xTxt.text()), 
                                                                       float(self.yTxt.text()), float(self.widthTxt.text()), 
                                                                       float(self.heightTxt.text()))


    def selectArea(self):
        pdfFileName = join(self.container.main_window.dbPath, Id2FileName(self.container.main_window.IdTxt.text())) + ".pdf"
        self.selectAreaDlg = PDFAreaSelector(pdfFileName)
        self.selectAreaDlg.areaSelected.connect(self.updateSelectedArea)
        self.selectAreaDlg.exec_()


    @pyqtSlot()
    def updateSelectedArea(self):
        self.noPageTxt.setText(str(self.selectAreaDlg.currentPageInd+1))
        self.xTxt.setText(str(self.selectAreaDlg.x))
        self.yTxt.setText(str(self.selectAreaDlg.y))
        self.widthTxt.setText(str(self.selectAreaDlg.width))
        self.heightTxt.setText(str(self.selectAreaDlg.height))

        w   = self.imgThumbnail.width()
        h   = self.imgThumbnail.height()
        self.imgThumbnail.setPixmap(self.selectAreaDlg.image.scaled(w, h, Qt.KeepAspectRatio))

    def loadThumbnail(self):
        pdfFileName = join(self.container.main_window.dbPath, Id2FileName(self.container.main_window.IdTxt.text())) + ".pdf"
        pixmap = loadImage(pdfFileName,
                          self.container.currentAnnotation.localizer.noPage,
                          self.container.currentAnnotation.localizer.x,
                          self.container.currentAnnotation.localizer.y,
                          self.container.currentAnnotation.localizer.width,
                          self.container.currentAnnotation.localizer.height)
        self.imgThumbnail.setPixmap(pixmap)




class EditAnnotTextWgt(QWidget):

    def __init__(self, container, parent=None):
        super().__init__(parent)

        self.container = container

        # Widgets
        self.localizeBtn         = QPushButton('Localize', self)
        self.correctBtn         = QPushButton('Correct', self)
        self.textToAnnotateTxt = QLineEdit('', self)
        self.contextTxt         = QLineEdit('', self)
        self.startTxt         = QLineEdit('', self)
    
        # Signals
        self.localizeBtn.clicked.connect(self.localizeText)
        self.textToAnnotateTxt.returnPressed.connect(self.localizeText)
        self.textToAnnotateTxt.textEdited.connect(self.setLocalizable)
        self.correctBtn.clicked.connect(self.correctText)

        # Layout
        self.editAnnotGroupBox = QGroupBox()
        gridAddAnnotations = QGridLayout(self.editAnnotGroupBox)
        gridAddAnnotations.addWidget(QLabel('Annotated text', self), 2, 0)
        gridAddAnnotations.addWidget(self.textToAnnotateTxt, 2, 1, 1, 2)
        gridAddAnnotations.addWidget(QLabel('Context', self), 3, 0)
        gridAddAnnotations.addWidget(self.contextTxt, 3, 1, 1, 2)

        gridAddAnnotations.addWidget(QLabel('Starting character', self), 4, 0)
        gridAddAnnotations.addWidget(self.startTxt, 4, 1)
        gridAddAnnotations.addWidget(self.correctBtn, 4, 2)
        gridAddAnnotations.addWidget(self.localizeBtn, 4, 3)

        self.setLayout(gridAddAnnotations)

        # Initial behavior
        self.editAnnotGroupBox.setDisabled(True)
        disableTextWidget(self.textToAnnotateTxt)
        disableTextWidget(self.contextTxt)
        disableTextWidget(self.startTxt)
        self.localizeBtn.setDisabled(True)
        self.correctBtn.setDisabled(True)


    def correctText(self):
        enableTextWidget(self.textToAnnotateTxt)
        self.localizeBtn.setDisabled(True)
        self.correctBtn.setDisabled(True)

        


    def localizeText(self):
        def recursiveSearch(queryString, text, a=0, level=0, maxLevel=5):

            starts = [(a, m.start(), len(queryString)) for m in re.finditer(re.escape(queryString), text)]
            if len(starts) == 0:
                if level < maxLevel and len(queryString) > 4:
                    N = len(queryString)
                    starts = []
                    starts.extend(recursiveSearch(queryString[:int(N/2)], text, a,             level+1, maxLevel))
                    starts.extend(recursiveSearch(queryString[int(N/2):], text, a+int(N/2), level+1, maxLevel))
                    return starts
                else:
                    return starts
            else:
                return starts

        def processBlocks(blocks):
            for block in blocks:
                matcher = SequenceMatcher(None, queryStr, fileText[block["start"]:block["end"]])
                ratio = matcher.ratio()

                matcher.set_seq2(fileText[block["start"]-1:block["end"]])
                new_ratio = matcher.ratio()
                while new_ratio >= ratio and block["start"] > 0: 
                    block["start"] -= 1
                    ratio = new_ratio
                    matcher.set_seq2(fileText[block["start"]-1:block["end"]])
                    new_ratio = matcher.ratio()

                matcher.set_seq2(fileText[block["start"]+1:block["end"]])
                new_ratio = matcher.ratio()
                while new_ratio >= ratio and block["start"] < len(fileText)-1: 
                    block["start"] += 1
                    ratio = new_ratio
                    matcher.set_seq2(fileText[block["start"]+1:block["end"]])
                    new_ratio = matcher.ratio()

                matcher.set_seq2(fileText[block["start"]:block["end"]-1])
                new_ratio = matcher.ratio()
                while new_ratio >= ratio and block["end"] > 0: 
                    block["end"] -= 1
                    ratio = new_ratio
                    matcher.set_seq2(fileText[block["start"]:block["end"]-1])
                    new_ratio = matcher.ratio()

                matcher.set_seq2(fileText[block["start"]:block["end"]+1])
                new_ratio = matcher.ratio()
                while new_ratio >= ratio and block["end"] < len(fileText)-1: 
                    block["end"] += 1
                    ratio = new_ratio
                    matcher.set_seq2(fileText[block["start"]:block["end"]+1])
                    new_ratio = matcher.ratio()


                block["candidate"]     = fileText[block["start"]:block["end"]]
                block["ratio"]         = ratio
                block["candidate"]     = block["candidate"].replace("\n", " ")        

            return np.array([block for block in blocks if block["ratio"] > 0.5])


        if len(self.textToAnnotateTxt.text()) < 3:
            errorMessage(self, "Error", "The text to localized should be at least 3-character long.")
            return            

        txtFileName = join(self.container.main_window.dbPath, Id2FileName(self.container.main_window.IdTxt.text())) + ".txt"
    
        with open(txtFileName, 'r', encoding="utf-8", errors='ignore') as f :
            fileText = f.read()

        ## We try to find an exact match...
        starts = [m.start() for m in re.finditer(re.escape(self.textToAnnotateTxt.text()), fileText)]

        ## If now exact match was found...        
        if len(starts) == 0:

            ## We try to find approximate matches using difflib
            queryStr = self.textToAnnotateTxt.text()
            N = len(self.textToAnnotateTxt.text())
            #s = SequenceMatcher(None, queryStr, fileText)
            #matches = [match for match in s.get_matching_blocks()]    
            blocks = []
            #for match in matches:
            #    block                = {}
            #    block["start"]        = max(0, match.b-match.a)
            #    block["end"]        = min(block["start"]+N, len(fileText))
            #    blocks.append(block)
            #blocks = processBlocks(blocks)


            ## No approximate match was found using difflib
            if len(blocks) == 0:

                ## We try to find partial matches using a recursive algorithm that sequentially
                ## splits the query string and try to find these subsetrings
                blocks = []
                for a, b, size in recursiveSearch(queryStr, fileText):
                    start = max(0, b-a)
                    blocks.append({"start":start,
                                   "end"  :min(start+N, len(fileText))})
                blocks = processBlocks(blocks)

            u, indices = np.unique([str(block["start"]) + "-" + str(block["end"]) for block in blocks], return_index=True)
            blocks = blocks[indices]
            blocks = sorted(blocks, key=lambda match: match["ratio"], reverse=True)
            matchDlg = MatchDlg(blocks, fileText, self)
            if matchDlg.exec_() == QDialog.Accepted:
                starts = [matchDlg.chosenBlock["start"]]
            else:
                return

        elif len(starts) > 1:
            blocks = [{"start":m.start(), 
                       "end":m.start()+len(self.textToAnnotateTxt.text()), 
                       "candidate":self.textToAnnotateTxt.text()} for m in re.finditer(self.textToAnnotateTxt.text(), fileText)]
            matchDlg = MatchDlg(blocks, fileText, self)
            if matchDlg.exec_() == QDialog.Accepted:
                starts = [matchDlg.chosenBlock["start"]]
            else:
                return
        
        contextStart = starts[0] - self.container.main_window.contextLength
        contextEnd = starts[0] + len(self.textToAnnotateTxt.text()) + self.container.main_window.contextLength
        self.contextTxt.setText(fileText[contextStart:contextEnd])
        
        localizer = TextLocalizer(self.textToAnnotateTxt.text(), starts[0])

        # FIXME A new Annotation shouldn't be created here!!
        self.container.currentAnnotation = Annotation(self.container.commentEdt.toPlainText(), 
                                                    [self.container.main_window.username],
                                                    self.container.main_window.IdTxt.text(), localizer)

        self.startTxt.setText(str(starts[0]))
        ID = self.container.main_window.IdTxt.text()
        if ID in self.container.main_window.selectedTagPersist:
            #self.currentAnnotation.clearTags()
            for tagId in self.container.main_window.selectedTagPersist[ID]:
                self.container.currentAnnotation.addTag(tagId, self.container.main_window.dicData[tagId])

        self.localizeBtn.setDisabled(True)
        self.correctBtn.setDisabled(False)
        disableTextWidget(self.textToAnnotateTxt)
        enableTextWidget(self.container.commentEdt)
        self.container.main_window.tagAnnotGroupBox.setDisabled(False)
        self.container.main_window.refreshTagList()
        self.container.newAnnotationBtn.setEnabled(True)
        self.container.duplicateAnnotationBtn.setEnabled(True)
        self.container.commentEdt.setFocus()
        self.container.main_window.detectAnnotChange = True
        self.container.main_window.needSaving = True

    def clearAnnotation(self):
        self.textToAnnotateTxt.setText("")
        self.contextTxt.setText("")
        self.startTxt.setText("")

    @pyqtSlot()
    def annotationSelectionChanged(self):

        if not self.container.currentAnnotation is None:
            if self.container.currentAnnotation.type == "text":
                self.textToAnnotateTxt.setText(self.container.currentAnnotation.text)
                self.contextTxt.setText(self.container.main_window.getCurrentContext())
                self.startTxt.setText(str(self.container.currentAnnotation.start))    
                self.localizeBtn.setEnabled(False) 
                self.correctBtn.setEnabled(True)
                disableTextWidget(self.textToAnnotateTxt)
        else:
            self.textToAnnotateTxt.setText("")
            self.contextTxt.setText("")    
            self.startTxt.setText("")    
            self.correctBtn.setEnabled(False)

    def setLocalizable(self):
        self.localizeBtn.setDisabled(len(self.textToAnnotateTxt.text()) < 3)


    def newAnnotation(self):
        enableTextWidget(self.textToAnnotateTxt)
        #self.textToAnnotateTxt.setFocus()



    def updateCurrentAnnotation(self):
        try:
            self.container.currentAnnotation.localizer = TextLocalizer(self.textToAnnotateTxt.text(), int(self.startTxt.text()))
        except ValueError:
            msgBox = QMessageBox(self)
            msgBox.setWindowTitle("Invalid localizer")
            msgBox.setText("Before saving changes to this annotation, you must enter an \"Annotated text\" and then localize it.")
            msgBox.setStandardButtons(QMessageBox.Ok)
            msgBox.exec_()



#!/usr/bin/python3


__author__ = "Christian O'Reilly"

from PyQt5.QtCore import QModelIndex, pyqtSignal, pyqtSlot, Qt, QAbstractTableModel
from PyQt5.QtWidgets import (QTableView, QMessageBox, QLabel, QGridLayout,
                             QAbstractItemView, QTextEdit, QWidget)

from nat.modelingParameter import getParameterTypes, getParameterTypeIDFromName
from nat.paramDesc import ParamDescPoint
from nat.parameterInstance import ParameterInstance
from nat.values import ValuesSimple, ValuesCompound
from nat.variable import NumericalVariable
from .itemDelegates import (ParamTypeCbo, DoubleDelegate, UnitDelegate,
                            StatisticsDelegate, ButtonDelegate)


parameterTypes = getParameterTypes()


class ParamValueWgt(QWidget):

    paramTypeSelected = pyqtSignal(str)

    def __init__(self, parent=None):
        super().__init__(parent)

        # FIXME Delayed refactoring.
        self._parent = parent

        # Widgets        
        self.paramsEdit = ParamTypeCbo(self)
        self.paramDescription = QTextEdit(self)
        self.valListTblWdg = ValueTableView()
        self.valListModel = ValueListModel(parent=self)
        self.valListTblWdg.setSelectionMode(QAbstractItemView.SingleSelection)
        self.valListTblWdg.setModel(self.valListModel)

        # Signals
        self.paramsEdit.activated[str].connect(self.paramTypeChanged)

        # Layout
        grid     = QGridLayout(self)
        grid.addWidget(QLabel("Parameter"), 0, 0)
        grid.addWidget(self.paramsEdit,           0, 1)
        grid.addWidget(self.paramDescription, 1, 0, 1, 2)
        grid.addWidget(self.valListTblWdg, 2, 0, 1, 2)

        # Initial behavior
        self.paramDescription.setReadOnly(True)

    @property
    def selectedTags(self):
        return self._parent.getSelectedTags()


    def paramTypeChanged(self, paramName):
        self.paramTypeSelected.emit(paramName)
        for paramType in parameterTypes:
            if paramType.name == paramName:
                self.paramDescription.setText(paramType.description)
            

    def newParameter(self):
        self.additionMode = True

        self.valListModel.clear()
        self.paramsEdit.setCurrentIndex(-1)
        self.paramDescription.setText("")
        self.paramsEdit.setFocus()


    def saveParameter(self, relationship, paramId):

        error = None
        typeID = getParameterTypeIDFromName(self.paramsEdit.currentText())
        if typeID is None:
            error = "parameter type"
        
        if error is None:
            values = NumericalVariable(typeID, self.valListModel.getValuesObject())
            description    = ParamDescPoint(values)
            return ParameterInstance(paramId, description, [], relationship)

        else:
            msgBox = QMessageBox(self)
            msgBox.setWindowTitle("Invalid modeling parameter")
            msgBox.setText("To save this annotation with an associated modeling parameter, a valid value and unit must be entered. The " + error + " entered is not valid.")
            msgBox.exec_() 
            return None

        
    def loadRow(self, currentParameter):
        if currentParameter is None:
            self.valListModel.clear()
            self.paramsEdit.setCurrentIndex(-1)
            self.paramDescription.setText("")
        else:
            self.valListModel.setFromParameter(currentParameter)    
            for i in range(self.paramsEdit.count()):
                if self.paramsEdit.itemText(i) == currentParameter.name:
                    self.paramsEdit.setCurrentIndex(i)
                    self.paramTypeChanged(currentParameter.name)
                    break


    def loadModelingParameter(self, row = None):
        if row is None:
            self.loadRow(None)
        else:
            self.loadRow(self._parent.currentAnnotation.parameters[row])


class ValueTableView(QTableView):

    def __init__(self, parent=None):
        super().__init__(parent)
        
        self.setItemDelegateForColumn(0, DoubleDelegate(self))
        self.setItemDelegateForColumn(1, UnitDelegate(self))
        self.setItemDelegateForColumn(2, StatisticsDelegate(self))
        self.setItemDelegateForColumn(3, ButtonDelegate(self))

    @pyqtSlot(bool)
    def cellButtonClicked(self, checked):
        # This slot will be called when our button is clicked. 
        # self.sender() returns a refence to the QPushButton created
        # by the delegate, not the delegate itself.
        self.model().deleteRow(self.sender().row)


class ValueListModel(QAbstractTableModel):

    def __init__(self, colHeader = ['Values', 'Unit', 'Statistic', 'delete'], parent=None):
        super().__init__(parent)
        #self.parameterList = parameterList
        self.clear(colHeader)

    def clear(self, colHeader = ['Values', 'Unit', 'Statistic', 'delete']):
        self.colHeader = colHeader
        self.values    = [] 

        self.refresh()



    def rowCount(self, parent=QModelIndex()):
        return len(self.values)+1

    def columnCount(self, parent=QModelIndex()):
        return len(self.colHeader)


    def getValues(self, row):
        return self.values[row].values

    def getUnit(self, row):
        return self.values[row].unit

    def getStatistic(self, row):
        return self.values[row].statistic

    
    def getDataByIndex(self, row, col):
        if row == self.rowCount()-1:
            return ""
        
        if col == 0:
            values = self.getValues(row)
            if len(values) == 1:
                return str(values[0])
            else:
                return str(values)
        elif col == 1:
            return self.getUnit(row)
        elif col == 2 :
            return self.getStatistic(row)
        else:
            return None


    def setDataByIndex(self, row, col, data):
        if row == self.rowCount()-1:
            self.values.append(ValuesSimple())
            self.refresh()
        
        if col == 0:
            values = eval(data)
            if isinstance(values, list):
                self.values[row].values    = values
            else:
                self.values[row].values    = [values]
                
        elif col == 1:
            self.values[row].unit      = data
        elif col == 2 :
            self.values[row].statistic = data
        else:
            return False
        return True


    def data(self, index, role=Qt.DisplayRole):
        if not index.isValid():
            return None

        if role != Qt.DisplayRole and  role != Qt.EditRole:
            return None

        return self.getDataByIndex(index.row(), index.column())



    def setData(self, index, value, role=Qt.DisplayRole):
        if value is None:
            value = ""
        return self.setDataByIndex(index.row(), index.column(), value)
        


    def flags(self, index):
        return Qt.ItemIsEditable | Qt.ItemIsEnabled



    def headerData(self, section, orientation, role=Qt.DisplayRole):
        if orientation == Qt.Horizontal and role == Qt.DisplayRole:
            return self.colHeader[section]        
        return None


    def refresh(self):
        self.layoutChanged.emit()

    def getValuesObject(self):
        """
         Return a ValuesSimple object if only one row has been filed, else
         return a ValuesCompound object containing all rows.
        """

        if self.rowCount() == 2:
            # Only one row is filed. An empty row is always appended to filed rows.
            return self.values[0]
        else:
            return ValuesCompound(self.values)
            
            
    def deleteRow(self, row):
        if 0 <= row < len(self.values):
            del self.values[row]
            self.refresh()

            
    def setFromParameter(self, parameter):
        if parameter is None:
            self.values = []
        elif isinstance(parameter.description.depVar, NumericalVariable):
            valuesObject = parameter.description.depVar.values
            if isinstance(valuesObject, ValuesSimple):
                self.values = [valuesObject]
            elif isinstance(valuesObject, ValuesCompound):
                self.values = valuesObject.valueLst
            else:
                raise TypeError
        else:
            raise TypeError

        self.refresh()
        
        
        


#!/usr/bin/python3

__author__ = "Christian O'Reilly"

from PyQt5.QtCore import pyqtSignal, Qt
from PyQt5.QtGui import QDoubleValidator
from PyQt5.QtWidgets import (QItemDelegate, QPushButton, QCheckBox,
                             QStyledItemDelegate, QLineEdit, QMessageBox,
                             QComboBox)

from nat.modelingParameter import getParameterTypes
from nat.values import unitIsValid, statisticList
from .autocomplete import AutoCompleteEdit


parameterTypes = getParameterTypes()


class ParamTypeCbo(AutoCompleteEdit):

    def __init__(self, parent=None):
        super().__init__(parent)
        self.setModel([paramType.name for paramType in parameterTypes])


class ButtonDelegate(QItemDelegate):
    """
    A delegate that places a fully functioning QPushButton in every
    cell of the column to which it's applied
    """

    def __init__(self, parent=None):
        # The parent is not an optional argument for the delegate as
        # we need to reference it in the paint method (see below)
        super().__init__(parent)
 
    def paint(self, painter, option, index):
        # This method will be called every time a particular cell is
        # in view and that view is changed in some way. We ask the 
        # delegates parent (in this case a table view) if the index
        # in question (the table cell) already has a widget associated 
        # with it. If not, create one with the text for this index and
        # connect its clicked signal to a slot in the parent view so 
        # we are notified when its used and can do something. 
        if not self.parent().indexWidget(index):
            button = QPushButton("delete", self.parent(), clicked=self.parent().cellButtonClicked)
            button.row = index.row()
            self.parent().setIndexWidget(index, button)


class CheckBoxDelegate(QItemDelegate):
    """
    A delegate that places a fully functioning QPushButton in every
    cell of the column to which it's applied
    """

    def __init__(self, parent=None):
        # The parent is not an optional argument for the delegate as
        # we need to reference it in the paint method (see below)
        super().__init__(parent)
 
    def paint(self, painter, option, index):
        # This method will be called every time a particular cell is
        # in view and that view is changed in some way. We ask the 
        # delegates parent (in this case a table view) if the index
        # in question (the table cell) already has a widget associated 
        # with it. If not, create one with the text for this index and
        # connect its clicked signal to a slot in the parent view so 
        # we are notified when its used and can do something. 
        if not self.parent().indexWidget(index):
            checkBox = QCheckBox(parent=self.parent(), clicked=self.parent().checkBoxClicked)
            checkBox.row = index.row()
            self.parent().setIndexWidget(index, checkBox)
        else:
            checkBox = self.parent().indexWidget(index)

        checkBox.setChecked(self.parent().model().data(index))




class DoubleDelegate(QStyledItemDelegate):

    def __init__(self, parent=None):
        super().__init__(parent)

    def createEditor(self, parent, option=None, index=None):
        edit = QLineEdit(parent)
        edit.setValidator(QDoubleValidator())
        return edit


class UnitDelegate(QStyledItemDelegate):

    def __init__(self, parent=None):
        super().__init__(parent)

    def createEditor(self, parent, option=None, index=None):
        return QLineEdit(parent=parent)

    def setModelData(self, editor, model, index):
        if unitIsValid(editor.text()):
            model.setData(index, editor.text()) #, QtGui.Qt.EditRole)
        else:
            msgBox = QMessageBox()
            msgBox.setWindowTitle("Invalid parameter unit")
            msgBox.setText("This unit is not valid.")
            msgBox.exec_()             


class ComboBoxDelegate(QStyledItemDelegate):

    typeSelected = pyqtSignal(str)

    def __init__(self, parent=None):
        super().__init__(parent)

    def createEditor(self, parent, option=None, index=None):
        self.comboBox = QComboBox(parent)
        self.comboBox.currentIndexChanged.connect(self.currentIndexChanged)
        return self.comboBox

    def setEditorData(self, editor, index):
        self.comboBox = editor
        editor.blockSignals(True)
        cboIndex = -1
        if not index.model().data(index) is None:
            for i in range(editor.count()):
                if editor.itemText(i) == index.model().data(index):
                    cboIndex = i
                    break
        editor.setCurrentIndex(cboIndex)    
        editor.blockSignals(False)



    def setModelData(self, editor, model, index):
        self.comboBox = editor
        model.setData(index, self.comboBox.currentText()) #, QtGui.Qt.EditRole)

    def updateEditorGeometry(self, editor, option, index):
        editor.setGeometry(option.rect)


    def currentIndexChanged(self):
        self.commitData.emit(self.sender())
        self.typeSelected.emit(self.comboBox.currentText())



class StatisticsDelegate(ComboBoxDelegate):

    def createEditor(self, parent, option=None, index=None):
        super().createEditor(parent, option, index)
        self.comboBox.addItems(statisticList)
        return self.comboBox


#class ExpPropertiesDelegate(ComboBoxDelegate):
#
#    def createEditor(self, parent, option=None, index=None):
#        super(ExpPropertiesDelegate, self).createEditor(parent, option, index)
#        self.comboBox.addItems(expPropertyStrList)
#        return self.comboBox


class ParamTypeDelegate(ComboBoxDelegate):

    def createEditor(self, parent, option=None, index=None):
        self.comboBox = ParamTypeCbo(parent)
        self.comboBox.currentIndexChanged.connect(self.currentIndexChanged)
        return self.comboBox


class AutoCompleteDelegate(QStyledItemDelegate):

    #typeSelected = QtCore.Signal(str)

    def __init__(self, parent=None):
        super().__init__(parent)

    def createEditor(self, parent, option=None, index=None):
        self.comboBox = AutoCompleteEdit(parent)
        #self.comboBox.currentIndexChanged.connect(self.currentIndexChanged)

        return self.comboBox

    """
    def setEditorData(self, editor, index):
        self.comboBox = editor
        editor.blockSignals(True)
        cboIndex = -1
        if not index.model().data(index) is None:
            for i in range(editor.count()):
                if editor.itemText(i) == index.model().data(index):
                    cboIndex = i
                    break
        editor.setCurrentIndex(cboIndex)    
        editor.blockSignals(False)
    """


    def setModelData(self, editor, model, index):
        self.comboBox = editor
        model.setData(index, self.comboBox.currentText()) #, QtGui.Qt.EditRole)

    def updateEditorGeometry(self, editor, option, index):
        editor.setGeometry(option.rect)



class ReqTagDelegate(AutoCompleteDelegate):

    cboNeedPopulation = pyqtSignal(str)

    def createEditor(self, parent, option=None, index=None):
        super().createEditor(parent, option, index)
        model = index.model() 
        row = index.row()
        index = model.index(row, 0)
        paramTypeName = model.data(index, Qt.UserRole)
        self.initialText = model.data(model.index(row, 1))
        self.cboNeedPopulation.emit(paramTypeName)    
        #self.comboBox.enterKeyPressed.connect(self.enterKeyPressed)

        return self.comboBox


    def addItems(self, items):
        self.comboBox.setModel(items)







# -*- coding: utf-8 -*-
"""
Created on Mon Sep  5 11:49:40 2016

@author: oreilly
"""

#!/usr/bin/python3

__author__ = "Christian O'Reilly"

import json

from PyQt5.QtWidgets import QDialog, QVBoxLayout, QTextEdit


class JSONDlg(QDialog):

    def __init__(self, parent=None):
        super().__init__(parent)

        self.setWindowTitle("JSON representations")
        self.setGeometry(100, 300, 1000, 1000)

        # Layout
        layout = QVBoxLayout()

        self.jsonText = QTextEdit(self)
        layout.addWidget(self.jsonText)

        self.setLayout(layout)

    def setJSON(self, objectToDisplay):
        self.jsonText.setText(json.dumps(objectToDisplay.toJSON(), 
                         sort_keys=True, indent=4, separators=(',', ': ')))        



__authors__ = ["Pierre-Alexandre Fonta", "Christian O'Reilly"]
__maintainer__ = "Pierre-Alexandre Fonta"

from collections import OrderedDict
from uuid import uuid4

from PyQt5.QtCore import QSize, pyqtSlot
from PyQt5.QtWidgets import (QDialog, QComboBox, QFormLayout, QLineEdit,
                             QStackedWidget, QDialogButtonBox, QTableWidget,
                             QWidget, QTableWidgetItem, QAbstractItemView,
                             QPushButton, QPlainTextEdit)

from neurocurator import utils


class ZoteroReferenceDialog(QDialog):

    UNPUBLISHED_ID_KEY = "UNPUBLISHED"
    UNPUBLISHED_ID_FIELD = "unpublishedId"
    ITEM_TYPE_FIELD = "itemType"
    CREATORS_FIELD = "creators"
    EXTRA_FIELD = "extra"

    def __init__(self, ref_templates, parent=None):
        super().__init__(parent)

        # Variables sections.

        # NB: Sorted.
        self._types = ["book", "bookSection", "conferencePaper", "document",
                       "forumPost", "journalArticle", "patent", "report",
                       "thesis", "webpage"]
        templates_subset = OrderedDict([(k, v) for k, v in ref_templates.items()
                                        if k in self._types])

        # Own configuration section.

        # TODO Better layout configuration.
        self.setMinimumSize(QSize(800, 1000))

        # Widgets section.

        # TODO Prevent edition of the reference type for an existing reference?
        self._types_edit = QComboBox()
        self._types_edit.addItems(self._types)

        self._types_widgets = self._templates_widgets(templates_subset)
        types_forms = self._templates_forms(self._types_widgets)
        forms_stack = QStackedWidget()
        for x in types_forms:
            forms_stack.addWidget(x)

        # NB: The first button with the accept role is made the default button.
        ok_cancel = QDialogButtonBox(QDialogButtonBox.Ok | QDialogButtonBox.Cancel)

        # Layouts section.

        layout = QFormLayout()
        layout.addRow(self.ITEM_TYPE_FIELD + ":", self._types_edit)
        layout.addRow(forms_stack)
        layout.addRow(ok_cancel)
        # NB: Don't use AllNonFixedFieldsGrow because it expands the QComboBox.
        layout.setFieldGrowthPolicy(QFormLayout.ExpandingFieldsGrow)
        utils.configure_form_layout(layout)
        self.setLayout(layout)

        # Signals section.

        self._types_edit.currentIndexChanged.connect(forms_stack.setCurrentIndex)
        ok_cancel.accepted.connect(self.accept)
        ok_cancel.rejected.connect(self.reject)

    # Data I/O methods section.

    def load_reference_data(self, ref_data):
        """Load the reference data into the corresponding edition form."""
        reference_type = ref_data[self.ITEM_TYPE_FIELD]
        self.select_reference_type(reference_type)
        type_widgets = self._types_widgets[reference_type]
        ignored_fields = []
        for field, widget in type_widgets.items():
            if field in ref_data:
                if field == self.CREATORS_FIELD:
                    widget.load_creators(ref_data[self.CREATORS_FIELD])
                elif field == self.EXTRA_FIELD:
                    # NB: '\n' is represented as a new line.
                    widget.setPlainText(ref_data[self.EXTRA_FIELD])
                else:
                    widget.setText(ref_data[field])
                    widget.home(False)
            elif field == self.UNPUBLISHED_ID_FIELD:
                if self.UNPUBLISHED_ID_KEY in ref_data[self.EXTRA_FIELD]:
                    widget.setDisabled(True)
            else:
                ignored_fields.append(field)
        # TODO Display in the GUI?
        if ignored_fields:
            print("Reference fields not loaded for edition: {}.".format(", ".join(ignored_fields)))

    def reference_data(self):
        """Get the reference data from the edition form."""
        reference_type = self._types_edit.currentText()
        type_widgets = self._types_widgets[reference_type]
        data = {self.ITEM_TYPE_FIELD: reference_type}
        for field, widget in type_widgets.items():
            if field == self.CREATORS_FIELD:
                headers = widget.HEADERS
                data[self.CREATORS_FIELD] = [{headers[0]: x[0], headers[1]: x[1], headers[2]: x[2]}
                                             for x in widget.creators()]
            elif field == self.EXTRA_FIELD:
                data[self.EXTRA_FIELD] = widget.toPlainText()
            elif field != self.UNPUBLISHED_ID_FIELD:
                # NB: Return an empty string by default.
                data[field] = widget.text()
        return data

    # Public methods section.

    def select_reference_type(self, ref_type):
        """Select the edition form for the reference type."""
        reference_type_index = self._types.index(ref_type)
        self._types_edit.setCurrentIndex(reference_type_index)

    # Private methods section.

    @staticmethod
    def _templates_forms(templates_widgets):
        """Return the edition forms for the dict {type: {field: widget}}."""
        forms = []
        for type_widgets in templates_widgets.values():
            form = QWidget()
            layout = QFormLayout()
            for field, widget in type_widgets.items():
                layout.addRow(field + ":", widget)
            layout.setFieldGrowthPolicy(QFormLayout.ExpandingFieldsGrow)
            utils.configure_form_layout(layout)
            form.setLayout(layout)
            forms.append(form)
        return forms

    def _templates_widgets(self, ref_templates):
        """Return the dict {type: {field: widget}} for the reference templates."""
        widgets = OrderedDict()
        special_fields = {self.ITEM_TYPE_FIELD, self.CREATORS_FIELD, self.EXTRA_FIELD}
        for ref_type, template in ref_templates.items():
            text_fields = [field for field, default in template.items()
                           if default == "" and field not in special_fields]
            # TODO Validate the input (QValidator, inputMask)?
            fields = OrderedDict()
            fields[self.CREATORS_FIELD] = CreatorsTableWidget()
            for x in text_fields:
                fields[x] = QLineEdit()
            fields[self.EXTRA_FIELD] = QPlainTextEdit()
            add_unpublished_id = QPushButton("Generate && Add in 'extra'")
            fields[self.UNPUBLISHED_ID_FIELD] = add_unpublished_id
            add_unpublished_id.clicked.connect(self._add_unpublished_id)
            widgets[ref_type] = fields
        return widgets

    def _add_unpublished_id(self):
        """Generate an UUID and add it as the UNPUBLISHED ID in the 'extra' field."""
        reference_type = self._types_edit.currentText()
        type_widgets = self._types_widgets[reference_type]
        type_widgets[self.UNPUBLISHED_ID_FIELD].setDisabled(True)
        # NB: uuid1() may compromise privacy since it creates a UUID containing
        # the computer’s network address. uuid4() creates a random UUID.
        uuid = str(uuid4())
        # NB: If the QPlainTextEdit is not empty, a new line is added before.
        # Otherwise, a new line is not added as first character.
        type_widgets[self.EXTRA_FIELD].appendPlainText("{}: {}".format(self.UNPUBLISHED_ID_KEY, uuid))
        # TODO After porting from Qt4 to Qt5, the disabling of the button and
        # the appended text are not painted. QWidget.update() and
        # qApp.processEvents() are not enough. The reason has not been found.
        self.repaint()


class CreatorsTableWidget(QTableWidget):

    HEADERS = ["firstName", "lastName", "creatorType"]

    def __init__(self, parent=None):
        # NB: AttributeError: 'columns()' is not a Qt property or a signal
        # if argument names are used (ie. rows, columns, parent).
        super().__init__(1, 3, parent)

        # Own configuration section.

        self.setCornerButtonEnabled(False)
        self.setHorizontalHeaderLabels(self.HEADERS)
        self.setSelectionBehavior(QAbstractItemView.SelectItems)
        self.setSelectionMode(QAbstractItemView.SingleSelection)
        self.verticalHeader().hide()
        # NB: Sorting is disabled by default.

        # Signals section.

        # TODO Instead, use a button / double click to add a new row?
        self.cellChanged.connect(self._add_row)

    # Slots section.

    @pyqtSlot(int, int)
    def _add_row(self, row, column):
        """Add an empty row at the end if the last row is not empty."""
        row_count = self.rowCount()
        if row == row_count - 1 and self.item(row, column).text() != "":
            self.insertRow(row_count)

    # Data I/O methods section.

    def load_creators(self, creators):
        """Load the creators data into the table."""
        row_count = len(creators)
        column_count = len(self.HEADERS)
        # NB: If sorting enabled, it may interfere with the insertion order.
        for i in range(row_count):
            creator = creators[i]
            self.insertRow(i)
            for j in range(column_count):
                cell_value = creator[self.HEADERS[j]]
                self.setItem(i, j, QTableWidgetItem(cell_value))

    def creators(self):
        """Get the creators data from the table as a nested list [rows: [columns: ]]."""
        row_range = range(self.rowCount() - 1)
        column_count = len(self.HEADERS)
        content = [[None] * column_count for _ in row_range]
        for i in row_range:
            for j in range(column_count):
                cell = self.item(i, j)
                if cell is not None:
                    content[i][j] = cell.text()
        return content



__authors__ = ["Pierre-Alexandre Fonta", "Christian O'Reilly"]
__maintainer__ = "Pierre-Alexandre Fonta"

import os

from setuptools import setup

NEUROCURATOR_VERSION = "0.4.4"
NAT_MINIMUM_VERSION = "0.4.2"

HERE = os.path.abspath(os.path.dirname(__file__))

# Get the long description from the README file.
# Convert to rst with: pandoc --from=markdown --to=rst README.md -o README.rst.
with open(os.path.join(HERE, "README.rst"), encoding="utf-8") as f:
    long_description = f.read()

setup(
    name="neurocurator",
    version=NEUROCURATOR_VERSION,
    description="Application to perform curation of neuroscientific literature.",
    long_description=long_description,
    keywords="neuroscience annotation curation literature modeling parameters",
    url="https://github.com/BlueBrain/neurocurator",
    author="Christian O'Reilly, Pierre-Alexandre Fonta",
    author_email="christian.oreilly@epfl.ch, pierre-alexandre.fonta@epfl.ch",
    # NB: 'If maintainer is provided, distutils lists it as the author in PKG-INFO'.
    # https://docs.python.org/3/distutils/setupscript.html#meta-data
    # maintainer="Pierre-Alexandre Fonta",
    # maintainer_email="pierre-alexandre@epfl.ch",
    license="GPLv3",
    packages=["neurocurator"],
    python_requires=">=3.4",
    install_requires=[
        "nat>=" + NAT_MINIMUM_VERSION,
        "PyQt5",
        "numpy",
        "pandas",
        "wand"
    ],
    data_files=[("", ["LICENSE.txt"])],
    entry_points={
        # NB: gui_scripts: on Windows no console is attached (no stdout/stderr).
        "console_scripts": ["neurocurator = neurocurator.__main__:main"]
    },
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering",
        "Environment :: X11 Applications :: Qt",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "License :: Free for non-commercial use",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Operating System :: MacOS",
        "Operating System :: POSIX :: Linux",
        "Operating System :: Microsoft :: Windows",
        "Natural Language :: English"
    ]
)



#!/usr/bin/python3

__author__ = "Christian O'Reilly"

from PyQt5.QtCore import pyqtSignal, pyqtSlot, QSize
from PyQt5.QtWidgets import (QLabel, QVBoxLayout, QAbstractItemView,
                             QListWidgetItem, QListWidget, QDialog, QTextEdit)


class TextEdit(QTextEdit):
    
    def __init__(self, no, parent=None):
        super().__init__(parent)
        self.no = no

    clicked = pyqtSignal(int)

    def mouseReleaseEvent(self, event):
        self.clicked.emit(self.no)


class MatchDlg(QDialog):

    def __init__(self, blocks, fileText, parent=None):
        super().__init__(parent)

        self.setWindowTitle("Approximate matches")
        self.setGeometry(100, 300, 1000, 200)

        NContext = 30

        self.listWidget = QListWidget(self)

        for row, block in enumerate(blocks):
            item = QListWidgetItem()

            before = fileText[block["start"]-NContext:block["start"]]
            after = fileText[block["end"]:block["end"]+NContext]

            item.setSizeHint(QSize(1000, 50))
            textEdit = TextEdit(row, self)
            textEdit.insertHtml('{}<b>{}</b>{}'.format(before, block["candidate"], after))
            self.listWidget.addItem(item)
            self.listWidget.setItemWidget(item, textEdit)

            textEdit.clicked.connect(self.selectText)

        self.listWidget.showMaximized()

        self.listWidget.setSelectionBehavior(QAbstractItemView.SelectRows)
        self.listWidget.setSelectionMode(QAbstractItemView.SingleSelection)

        layout = QVBoxLayout()
        layout.addWidget(QLabel('The text to annotate has not been found. Similar entries are listed below. Please choose correct one.', self))
        layout.addWidget(self.listWidget)
        self.setLayout(layout)

        self.blocks = blocks
        self.chosenBlock = None

    @pyqtSlot(int)
    def selectText(self, row):
        self.chosenBlock = self.blocks[row]
        self.accept()



#!/usr/bin/python3

__author__ = 'oreilly'
__email__  = 'christian.oreilly@epfl.ch'

from PyQt5.QtCore import Qt, QModelIndex, QAbstractItemModel
from PyQt5.QtWidgets import QTreeView


class TreeModel(QAbstractItemModel):

    def __init__(self, tree, parent=None):
        super().__init__(parent)

        self.__tree = tree

        if len(tree):
            self.__current = tree[0]

    def flags(self, index):
        flag = Qt.ItemIsEnabled
        if index.isValid():
            flag |= Qt.ItemIsSelectable
        return flag


    def index(self, row, column, parent=QModelIndex()):
        if parent.isValid():
            node = parent.internalPointer().children[row]
        else:
            if not len(self.__tree):
                return
            node = self.__tree[row]
        return self.__createIndex(row, column, node)


    def parent(self, index):
        node = QModelIndex()
        if index.isValid():
            nodeS = index.internalPointer()
            parent = nodeS.parent
            if parent is not None:
                node = self.__createIndex(parent.position(), 0, parent)
        return node


    def rowCount(self, parent=QModelIndex()):
        count = len(self.__tree)
        node = parent.internalPointer()
        if node is not None:
            count = len(node.children)
        return count


    def columnCount(self, parent=QModelIndex()):
        return 1


    def data(self, index, role=Qt.DisplayRole):
        data = None
        if role == Qt.DisplayRole or role == Qt.EditRole:
            node = index.internalPointer()
            data = node.txt

        if role == Qt.UserRole:
            node = index.internalPointer()
            data = node.id

        return data


    def setData(self, index, value, role=Qt.DisplayRole):
        result = True
        if role == Qt.EditRole and value != "":
            node = index.internalPointer()
            node.text = value
            result = True
        return result


    def __createIndex(self, row, column, node):

        if hasattr(node, "index"):
            if not node.index is None: 
                return node.index
        
        index = self.createIndex(row, column, node)
        node.index = index
        return node.index

    def headerData(self, section, orientation, role=Qt.DisplayRole):
        return None





class TreeView(QTreeView):

    def __init__(self, model, parent=None):
        super().__init__(parent)

        self.__model = model
        self.setModel(model)

        self.header().hide()

        index = self.__model.index(0, 0)
        if not index is None:
            self.setCurrentIndex(index)



# -*- coding: utf-8 -*-
"""
Created on Mon Sep  5 11:49:40 2016

@author: oreilly
"""

#!/usr/bin/python3

__author__ = "Christian O'Reilly"

from PyQt5.QtWidgets import QDialog, QLabel, QGroupBox, QVBoxLayout, QHBoxLayout


class AddOntoTermDlg(QDialog):

    def __init__(self, parent=None):
        super().__init__(parent)

        self.setWindowTitle("Addition of a new ontology term")
        self.setGeometry(100, 300, 1000, 1000)

        self.warningMsg = QLabel("Remember: ", self)

        self.addingGroup = QGroupBox("Adding a new ontology term")
        addingLayout = QVBoxLayout(self.addingGroup)
        addingLayout.addWidget(self.warningMsg)

        self.suggestionGroup = QGroupBox("Existing ontological term suggestions")

        # Layout
        layout = QHBoxLayout()
        layout.addWidget(self.addingGroup)
        layout.addWidget(self.suggestionGroup)

        self.setLayout(layout)



#!/usr/bin/python3

__author__ = "Christian O'Reilly"

from PyQt5.QtWidgets import QLabel, QGridLayout, QComboBox, QWidget, QStackedWidget

from nat.relationship import Relationship
from nat.tag import Tag


class ParamRelationWgt(QWidget):

    def __init__(self, parent=None):
        super().__init__(parent)

        # FIXME Delayed refactoring.
        self._parent = parent

        # Widgets        
        self.relationshipCbo    = QComboBox(self)
        self.relationshipCbo.addItems(["unspecified", "point", "directed", "undirected"])

        self.relationStack        = QStackedWidget(self)

        self.relationWgt   = QWidget()
        self.relEntity1Cbo = QComboBox(self.relationWgt)
        self.relEntity2Cbo = QComboBox(self.relationWgt)
        self.relEntity1Lbl = QLabel("Entity 1")
        self.relEntity2Lbl = QLabel("Entity 2")
        self.infoLabel     = QLabel("Note: Entity comboboxes are populated from selected tags.")

        relationLayout = QGridLayout(self.relationWgt)
        relationLayout.addWidget(self.relEntity1Lbl, 0, 0)
        relationLayout.addWidget(self.relEntity2Lbl, 1, 0)
        relationLayout.addWidget(self.relEntity1Cbo, 0, 1)
        relationLayout.addWidget(self.relEntity2Cbo, 1, 1)
        relationLayout.addWidget(self.infoLabel, 2, 0, 1, 2)
        
        # Signals
        self.relationshipCbo.currentIndexChanged.connect(self.relationSelected)

        # Layout
        grid     = QGridLayout(self)

        grid.addWidget(QLabel("Relationship"),  0, 0)
        grid.addWidget(self.relationshipCbo,  0, 1)
        grid.addWidget(self.relationWgt,  1, 0, 1, 2)

        # Initial behavior
        self.refreshEntityCbo()
        self.relationSelected(0)

    def clear(self):
        self.refreshEntityCbo()
        self.relationSelected(0)



    @property
    def selectedTags(self):
        return self._parent.getSelectedTags()


    def refreshEntityCbo(self):
        selectedTagNames  = [tag.name for tag in self.selectedTags]
        self.relEntity1Cbo.clear()
        self.relEntity2Cbo.clear()
        self.relEntity1Cbo.addItems(selectedTagNames)
        self.relEntity2Cbo.addItems(selectedTagNames)


    def relationSelected(self, index):
        
        if self.relationshipCbo.currentText() == "unspecified":
            self.relEntity2Cbo.setVisible(False)
            self.relEntity1Cbo.setVisible(False)
            self.relEntity1Lbl.setVisible(False)
            self.relEntity2Lbl.setVisible(False)
            self.infoLabel.setVisible(False)
        elif self.relationshipCbo.currentText() == "point":
            self.relEntity2Cbo.setVisible(False)
            self.relEntity1Cbo.setVisible(True)
            self.relEntity1Lbl.setVisible(True)
            self.relEntity2Lbl.setVisible(False)
            self.infoLabel.setVisible(True)                        
            self.relEntity1Lbl.setText("Entity")
        elif self.relationshipCbo.currentText() == "directed":
            self.relEntity2Cbo.setVisible(True)
            self.relEntity1Cbo.setVisible(True)
            self.relEntity1Lbl.setVisible(True)
            self.relEntity2Lbl.setVisible(True)
            self.infoLabel.setVisible(True)  
            self.relEntity1Lbl.setText("Entity from")
            self.relEntity2Lbl.setText("Entity to")
        elif self.relationshipCbo.currentText() == "undirected":
            self.relEntity2Cbo.setVisible(True)
            self.relEntity1Cbo.setVisible(True)
            self.relEntity1Lbl.setVisible(True)
            self.relEntity2Lbl.setVisible(True)
            self.infoLabel.setVisible(True)  
            self.relEntity1Lbl.setText("Entity 1")
            self.relEntity2Lbl.setText("Entity 2")
        else:
            raise ValueError


    def getRelationship(self):

        if self.relationshipCbo.currentText() == "unspecified":
            return None

        for tag in self.selectedTags:
            if tag.name == self.relEntity1Cbo.currentText():
                tag1 = Tag(tag.id, tag.name)
                break

        if self.relationshipCbo.currentText() == "point":
            tag2 = None
        else:
            for tag in self.selectedTags:
                if tag.name == self.relEntity2Cbo.currentText():
                    tag2 = Tag(tag.id, tag.name)
                    break

        return Relationship(self.relationshipCbo.currentText(), tag1, tag2)


    def setRelationship(self, relationship):

        if relationship is None:
            self.relationshipCbo.setCurrentIndex(0)
        else:    
            for noRel, relationType in enumerate(["point", "directed", "undirected"]):
                if relationType == relationship.type:
                    self.relationshipCbo.setCurrentIndex(noRel+1)
                    break

        if self.relationshipCbo.currentText() != "unspecified":
            for i in range(self.relEntity1Cbo.count()):
                if self.relEntity1Cbo.itemText(i) == relationship.entity1.name:
                    self.relEntity1Cbo.setCurrentIndex(i)
                    break

        if not self.relationshipCbo.currentText() in ["unspecified", "point"]:
            for i in range(self.relEntity2Cbo.count()):
                if self.relEntity2Cbo.itemText(i) == relationship.entity2.name:
                    self.relEntity2Cbo.setCurrentIndex(i)
                    break
        


    def loadRow(self, currentParameter):
        """
         Called when a row has been selected in the table listing all the modeling parameters.
         It update the interface with the values associated with this specific parameter.
        """
        self.setRelationship(currentParameter.relationship)


    def loadModelingParameter(self, row = None):
        """
         Call when a new annotation has been selected so that all the modeling parameters
         associated with this annotation are loaded in the parameter list. 
        """
        self.refreshEntityCbo()




